/*
**  File: traffic.c
**
**  Author: Edward Chow
**
**  Comments: collecting traffic volume using pcap
**            and report traffic volume to a plotting program via datagram 
**            socket.
**
*/

#include <stdio.h>
#include <pcap.h>
#include <unistd.h>
#include <netinet/in.h>
#include "ethernet.h"
#include <sys/time.h>
#include <sys/shm.h>
#include <sys/sem.h>

#define SNAPLEN      14
#define PROMISC      1
#define READ_TIMEOUT 2000

typedef void (*pcap_func_t)(u_char *, const struct pcap_pkthdr *, 
			    const u_char *);


struct timeval stv, etv;
struct timezone tz;

pcap_t  *pd;               /* pcap device descriptor */
int     cnt;               /* number of packets to read */
pcap_handler dev_prcsr;    /* ptr to func that processes packet for a device */
char errorbuf[PCAP_ERRBUF_SIZE];
bpf_u_int32 localnet, netmask;
struct bpf_program fcode;
int debug = 0;
double traffic;
double totalBytes;
double totalBytesLastPeriod;
int reportPeriod = 2000000; /* set at 2 seconds */

/* the following shared memory code is in Advanced Unix Programming p. 194 */

#define MAXMSG 4096
#define MAXOPEN 20
#define BADADDR (char *) (-1)

int sndsid, rcvsid, segid;
char *addr;

static int findinfo(key, sndsidp, rcvsidp, addrp, segidp)
int key;
int *sndsidp, *rcvsidp, *segidp;
char **addrp;
{

  static struct {
   int segid;  /* segment id */
   int key;    /* key        */
   char *addr; /* addr of shared memory segment */
   int sndsid; /* semaphore-ID for writing totalBytes */
   int rcvsid;  /* semaphore-ID for reading totalBytes */
  } sems[MAXOPEN];
  int i, avail;
  extern int errno;
  char *shmat();
  
  avail = -1;
  for (i=0; i<MAXOPEN; i++) {
    if (sems[i].key == key) {
      *sndsidp = sems[i].sndsid;
      *rcvsidp = sems[i].rcvsid;
      *segidp = sems[i].segid;
      *addrp = sems[i].addr;
      return(1);
    }
    if (sems[i].key == 0 && avail == -1)
      avail = i;
  }
  if (avail == -1) {
    errno=0;
    return(0);
  }
  sems[avail].key = key;
/*
  *sndsidp = sems[avail].sndsid = semtrans(key);
  *rcvsidp = sems[avail].rcvsid = semtrans(key+10000);
*/
  if ((*segidp=sems[avail].segid=shmget((key_t)key, MAXMSG, 0666|IPC_CREAT))==-1) {
    printf("shmget fail\n");
    return (0);
  }
  if ((*addrp=sems[avail].addr=shmat(*segidp, 0, 0))==BADADDR)  {
    printf("shmat fail\n");
    return (0);
  }
  return (1);
}

void dump_ethernet(u_char *user, const struct pcap_pkthdr *h, u_char *p)
{
  int length; 
  int caplen;
  int ether_type;
  EtherHdr *ep;
  char *etheraddr_string(u_char *);
  char *etherproto_string(u_short);
  int time;

  /*
   * Get total packet length and length of the captured section
   */

  if (debug) fprintf(stderr, "get in dumpethernet\n");
  length = h->len;
  caplen = h->caplen;
  totalBytes += length+4.0; /* add check 4 byte FCS */
  /* P(sndsid); need to add semaphor maybe */
  memcpy(addr, (char *) &totalBytes,  8);  /* set shared memory variable */
  /* V(sndsid); */

/*  let fork child process handle the reporting  
  (void) gettimeofday(&etv, &tz);
  time = (etv.tv_sec-stv.tv_sec)*1000000+(etv.tv_usec-stv.tv_usec);
  if (time > reportPeriod -4) {
     printf("time=%d usec, totalBytes=%f, traffic=%f kbps\n", 
            time, totalBytes, totalBytes*8*1000.0/time);
     (void) gettimeofday(&stv, &tz); 
     totalBytes = 0;
  }
*/
  /*
   * Dump header announcement
   */
 
 if (debug) {

  printf("=================================================================\n");
  printf("                        Ethernet Header(%u.%06u)\n",
	 (u_int32_t) h->ts.tv_sec, (u_int32_t) h->ts.tv_usec);
  printf("-----------------------------------------------------------------\n");
}

  /*
   * Check for truncated header 
   */

  if (caplen < sizeof(EtherHdr)) 
    {
      printf("Ethernet header too short! (%d bytes)\n", length);
      return;
    }


  /*
   * Dump header fields
   */

 if (debug) {
  ep = (EtherHdr *) p;  
  ether_type = ntohs(ep->ether_type);

  printf("Hardware source:        %s\n", 
	     etheraddr_string(ep->ether_src));
  printf("Hardware destination:   %s\n", 
	     etheraddr_string(ep->ether_dst));
  printf("Protocol type:          %xH (%s)\n", ether_type, 
	     etherproto_string(ep->ether_type));
  printf("Length:                 %d\n", length+4); /* add FCS */
  printf("totalLength:            %f\n", totalBytes);
 }
  
}

int main(int argc, char* argv[]){
  u_char *userdata;
  int time;
  double totalTraffic;

  userdata = NULL;
  cnt = 0;


  (void) gettimeofday(&stv, &tz);
  pd = pcap_open_live("eth0", SNAPLEN, PROMISC, READ_TIMEOUT, errorbuf);
  if (pd == NULL) {
    printf("open_pcap: pcap_open_live() failed for %s: %s", 
		    "eth0", errorbuf);
    exit(1);
  }
 
  /*
   * Compile command line filter spec info fcode FSM
   */

  if (pcap_lookupnet("eth0", &localnet, &netmask, errorbuf) < 0) {
    printf("open_pcap: pcap_lookupnet() failed for eth0 %s\n", errorbuf);
    exit(1);
  }
  
  if (pcap_compile(pd, &fcode, "", 0, netmask) < 0){
    printf("pcap_compile: %s", pcap_geterr(pd)); 
    exit(1);
  }

  /*
   * Set the pcap filter with our fcode FSM.  That should do it...
   */

  if (pcap_setfilter(pd, &fcode) < 0) {
    printf("pcap_setfilter: %s", pcap_geterr(pd));
    exit(1);
  }

  setuid(getuid());
  init_addrtoname();
  dev_prcsr = (pcap_func_t) dump_ethernet;

  (void) gettimeofday(&etv, &tz);
  time = (etv.tv_sec-stv.tv_sec)*1000000+(etv.tv_usec-stv.tv_usec);
  printf("pcap setup time=%d usec, stv=%d.%d, etv=%d.%d\n", time, stv.tv_sec, stv.tv_usec,
          etv.tv_sec, etv.tv_usec);
  printf("entering pcap_loop\n");
  totalBytes = 0;
  totalBytesLastPeriod = 0;
  cnt = 0;
  (void) gettimeofday(&stv, &tz);

  if (!findinfo(10, &sndsid, &rcvsid, &addr, &segid)) 
    return(1);
  /* P(sndsid); */
  memcpy(addr, (char *) &totalBytes,  8);  /* set shared memory variable */
  /* V(sndsid); */
 
  if (fork()== 0) {
    while (1) {
      (void) gettimeofday(&etv, &tz);
      time = (etv.tv_sec-stv.tv_sec)*1000000+(etv.tv_usec-stv.tv_usec);
      if (time >= reportPeriod) {
        /* P(sndsid); */
        memcpy((char *) &totalBytes, addr, 8);  /* get shared memory variable */
        /* V(sndsid); */
        traffic = totalBytes -totalBytesLastPeriod;
 
        printf("fork time=%d usec, totalBytes=%f, traffic=%f kbps\n", 
            time, traffic, traffic*8*1000.0/time);
        totalBytesLastPeriod = totalBytes;
       (void) gettimeofday(&stv, &tz); /* start new time period */

        /* the following does not seem to work 
          totalBytes = 0;
          memcpy(addr, (char *) &totalBytes,  4);  
        */
      }
    }
  } else {
    if (pcap_loop(pd, cnt, dev_prcsr, userdata) < 0){
      printf("pcap_loop: %s", pcap_geterr(pd));
      exit(1);
    }
  }
  /*
   * Close the pcap device
   */
  

  pcap_close(pd);

  exit(0);
}

