/*
**  File: traffic.c
**
**  Author: Edward Chow
**
**  Comments: collecting traffic volume using pcap
**            and report traffic volume to a plotting program via datagram 
**            socket.
**
*/

#include <stdio.h>
#include <pcap.h>
#include <unistd.h>
#include <netinet/in.h>
#include "ethernet.h"
#include <sys/time.h>

#define SNAPLEN      14
#define PROMISC      1
#define READ_TIMEOUT 2000

typedef void (*pcap_func_t)(u_char *, const struct pcap_pkthdr *, 
			    const u_char *);

struct timeval stv, etv;
struct timezone tz;

pcap_t  *pd;               /* pcap device descriptor */
int     cnt;               /* number of packets to read */
pcap_handler dev_prcsr;    /* ptr to func that processes packet for a device */
char errorbuf[PCAP_ERRBUF_SIZE];
bpf_u_int32 localnet, netmask;
struct bpf_program fcode;
int debug = 0;
int totalBytes;
int reportPeriod = 2000000; /* set at 2 seconds */

void dump_ethernet(u_char *user, const struct pcap_pkthdr *h, u_char *p)
{
  int length; 
  int caplen;
  int ether_type;
  EtherHdr *ep;
  char *etheraddr_string(u_char *);
  char *etherproto_string(u_short);
  int time;

  /*
   * Get total packet length and length of the captured section
   */

  if (debug) fprintf(stderr, "get in dumpethernet\n");
  length = h->len;
  caplen = h->caplen;
  totalBytes += length+4; /* add check 4 byte FCS */

  (void) gettimeofday(&etv, &tz);
  time = (etv.tv_sec-stv.tv_sec)*1000000+(etv.tv_usec-stv.tv_usec);
  if (time > reportPeriod -4) {
     printf("time=%d usec, totalBytes=%d, traffic=%f kbps\n", 
            time, totalBytes, totalBytes*8*1000.0/time);
     (void) gettimeofday(&stv, &tz); /* start new time period */
     totalBytes = 0;
  }

  /*
   * Dump header announcement
   */
 
 if (debug) {

  printf("=================================================================\n");
  printf("                        Ethernet Header(%u.%06u)\n",
	 (u_int32_t) h->ts.tv_sec, (u_int32_t) h->ts.tv_usec);
  printf("-----------------------------------------------------------------\n");
}

  /*
   * Check for truncated header 
   */

  if (caplen < sizeof(EtherHdr)) 
    {
      printf("Ethernet header too short! (%d bytes)\n", length);
      return;
    }


  /*
   * Dump header fields
   */

 if (debug) {
  ep = (EtherHdr *) p;  
  ether_type = ntohs(ep->ether_type);

  printf("Hardware source:        %s\n", 
	     etheraddr_string(ep->ether_src));
  printf("Hardware destination:   %s\n", 
	     etheraddr_string(ep->ether_dst));
  printf("Protocol type:          %xH (%s)\n", ether_type, 
	     etherproto_string(ep->ether_type));
  printf("Length:                 %d\n", length+4); /* add FCS */
  printf("totalLength:            %d\n", totalBytes);
 }
  
}

int main(int argc, char* argv[]){
  u_char *userdata;
  int time;

  userdata = NULL;
  cnt = 0;


  (void) gettimeofday(&stv, &tz);
  pd = pcap_open_live("eth0", SNAPLEN, PROMISC, READ_TIMEOUT, errorbuf);
  if (pd == NULL) {
    printf("open_pcap: pcap_open_live() failed for %s: %s", 
		    "eth0", errorbuf);
    exit(1);
  }
 
  /*
   * Compile command line filter spec info fcode FSM
   */

  if (pcap_lookupnet("eth0", &localnet, &netmask, errorbuf) < 0) {
    printf("open_pcap: pcap_lookupnet() failed for eth0 %s\n", errorbuf);
    exit(1);
  }
  
  if (pcap_compile(pd, &fcode, "", 0, netmask) < 0){
    printf("pcap_compile: %s", pcap_geterr(pd)); 
    exit(1);
  }

  /*
   * Set the pcap filter with our fcode FSM.  That should do it...
   */

  if (pcap_setfilter(pd, &fcode) < 0) {
    printf("pcap_setfilter: %s", pcap_geterr(pd));
    exit(1);
  }

  setuid(getuid());
  init_addrtoname();
  dev_prcsr = (pcap_func_t) dump_ethernet;

  (void) gettimeofday(&etv, &tz);
  time = (etv.tv_sec-stv.tv_sec)*1000000+(etv.tv_usec-stv.tv_usec);
  printf("pcap setup time=%d usec, stv=%d.%d, etv=%d.%d\n", time, stv.tv_sec, stv.tv_usec,
          etv.tv_sec, etv.tv_usec);
  printf("entering pcap_loop\n");
  totalBytes = 0;
  cnt = 0;
  (void) gettimeofday(&stv, &tz);

  if (pcap_loop(pd, cnt, dev_prcsr, userdata) < 0){
      printf("pcap_loop: %s", pcap_geterr(pd));
      exit(1);
  }
  /*
   * Close the pcap device
   */
  

  pcap_close(pd);

  exit(0);
}

